use super::uni_backup_task::*;
use cyfs_backup_lib::*;
use cyfs_base::*;
use cyfs_bdt::ChunkReaderRef;
use cyfs_lib::*;

use std::sync::{Arc, Mutex};

pub struct BackupManager {
    noc: NamedObjectCacheRef,
    ndc: NamedDataCacheRef,
    chunk_reader: ChunkReaderRef,

    // all backup tasks
    tasks: Mutex<Vec<UniBackupTask>>,
}

impl BackupManager {
    pub fn new(
        noc: NamedObjectCacheRef,
        ndc: NamedDataCacheRef,
        chunk_reader: ChunkReaderRef,
    ) -> Self {
        
        Self {
            noc,
            ndc,
            chunk_reader,

            tasks: Mutex::new(vec![]),
        }
    }

    fn create_uni_backup_task(&self, params: &UniBackupParams) -> BuckyResult<UniBackupTask> {
        let task = UniBackupTask::new(
            params.id.clone(),
            self.noc.clone(),
            self.ndc.clone(),
            self.chunk_reader.clone(),
        );

        {
            let mut tasks = self.tasks.lock().unwrap();
            if tasks.iter().find(|item| item.id() == params.id).is_some() {
                let msg = format!("backup task already exists! task={}", params.id);
                error!("{}", msg);
                return Err(BuckyError::new(BuckyErrorCode::AlreadyExists, msg));
            }

            tasks.push(task.clone());
        }

        Ok(task)
    }

    pub async fn run_uni_backup(&self, params: UniBackupParams) -> BuckyResult<()> {
        let task = self.create_uni_backup_task(&params)?;

        task.run(params).await
    }

    pub async fn start_uni_backup(&self, params: UniBackupParams) -> BuckyResult<()> {
        let task = self.create_uni_backup_task(&params)?;

        async_std::task::spawn(async move {
            let id = params.id.clone();
            match task.run(params).await {
                Ok(()) => {
                    info!("run uni backup task complete! task={}", id);
                }
                Err(e) => {
                    error!("run uni backup task failed! task={}, {}", id, e);
                }
            }
        });

        Ok(())
    }

    pub fn get_task_status(&self, id: &str) -> BuckyResult<BackupStatus> {
        let status = {
            let tasks = self.tasks.lock().unwrap();
            let ret = tasks.iter().find(|item| item.id() == id);
            if ret.is_none() {
                let msg = format!("backup task not exists! task={}", id);
                error!("{}", msg);
                return Err(BuckyError::new(BuckyErrorCode::NotFound, msg));
            }

            ret.unwrap().status()
        };

        Ok(status)
    }
}

pub type BackupManagerRef = Arc<BackupManager>;
